/*
 * Created on Jul, 2007, revised 2008-2009
 *
 */
package org.mindswap.pellet;

import aterm.*;
import org.mindswap.pellet.exceptions.*;
import org.mindswap.pellet.tbox.*;
import org.mindswap.pellet.utils.*;
import org.mindswap.pellet.utils.Timer;

import java.util.*;

/**
 * @author NataKeberle
 */

public class OWLMETStrategy extends CompletionStrategy {
	 /**
     * @param abox
     * @param blocking
     */
//    protected TemporalMarking tblocking; - currently expanding domain - so don't use temporal marks
    private LinkedList mayNeedExpanding;

	IntMap ts;
    IntMap tsInducted; //TODO for non-hybrid expansion of temporal formula, when TimeStructure is empty and new moments are created only from Fn/Pn rules
	TBox tbox;
//     ATermList conj;
    Individual root;
//TODO added for combined TSAT+SAT
    public HashMap OWLConcepts;
    KnowledgeBase kb;

    public OWLMETStrategy(ABox abox) {
		super(abox, new SubsetBlocking());
    }

    public void initialize() {
//       ABox.DEBUG = true;
        super.initialize();
        kb = abox.getKB();
        tbox = abox.getKB().getTBox();
        ts = abox.getKB().ts;
        tsInducted = ts;

//TODO added for combined  TSAT+SAT
        OWLConcepts = new HashMap();

        root = (Individual) abox.getNodes().iterator().next();

        unfoldingMap = tbox.getUnfoldingMap();

        for(Iterator i = abox.getBranches().iterator(); i.hasNext();) {
            Branch branch = (Branch) i.next();
            branch.setStrategy(this);
        }
		return;
	}

    ABox complete() {
        Timer t;
        completionTimer.start();
        Expressivity expressivity = abox.getKB().getExpressivity();
        boolean fullDatatypeReasoning =
            PelletOptions.USE_FULL_DATATYPE_REASONING &&
            (expressivity.hasCardinalityD() || expressivity.hasKeys());
        
        initialize();

        mayNeedExpanding = new LinkedList();
        mayNeedExpanding.add( root );

        while(!abox.isComplete()) {
            while(abox.changed && !abox.isClosed()) {
                completionTimer.check();

                abox.changed = false;
               
                if(ABox.DEBUG) {
                    System.out.println("Branch: " + abox.getBranch() +
                        ", Depth: " + abox.treeDepth + ", Size: " + abox.getNodes().size() + 
                        ", Mem: " + (Runtime.getRuntime().freeMemory()/1000) + "kb");
                    abox.validate();
                    printBlocked();
                    abox.printTree();
                }

                applyARule(); //TODO every individual passes through applyARule() in expand(ind)
                Individual ind = getNextIndividual();
                if (ABox.DEBUG)
                    System.out.println("Expansion start for individual:   " + ind.toString());
				expand(ind);
//TODO  added for combined TSAT+SAT : createOWLConcept
                if( PelletOptions.COMBINED_TSAT_SAT ) {
                    ATermAppl moment = createMomentConcept(ind);
                    if (moment.getName()!="now"){
                        ATermAppl new_c = createOWLConcept(ind);
                        if (new_c!=null)
                            OWLConcepts.put(moment,new_c);
                        if (!kb.constructedOWL.containsKey(moment) && !kb.constructedOWL.containsValue(new_c))
                            kb.constructedOWL.put(moment, new_c);
                    }

//TODO added for combined TSAT+SAT :  createOWLConcept
                }
            }
            if( abox.isClosed() ) {
                if(ABox.DEBUG)
                    System.out.println(
                        "Clash at Branch (" + abox.getBranch() + ") " + abox.getClash());

//                if(backtrack())
//                    abox.setClash( null );
//                else
                    abox.setComplete( true );
            }
            else {
            	if (PelletOptions.SATURATE_TABLEAU) {
            		Branch unexploredBranch = null;
                	for (int i=abox.getBranches().size()-1; i>=0; i--) {
                		unexploredBranch = (Branch) abox.getBranches().get(i);
                        unexploredBranch.tryNext++;
                		if (unexploredBranch.tryNext < unexploredBranch.tryCount) {
                			restore(unexploredBranch);
                            System.out.println("restoring branch "+ unexploredBranch.branch + " tryNext = "+unexploredBranch.tryNext + " tryCount = "+unexploredBranch.tryCount);
                			unexploredBranch.tryNext();
                			break;
                		}
                		else { 
                            System.out.println("removing branch "+ unexploredBranch.branch);
                            abox.getBranches().remove(i);
                            unexploredBranch = null;
                        }
                	}
                	if(unexploredBranch == null) {
                		abox.setComplete( true );	
                	}
            	}
                else abox.setComplete( true );
            }
        }
        completionTimer.stop();
        if (ABox.DEBUG) System.out.println("RECONSTRUCTED CONCEPT FOR SAT CHECKING IS   :    "+ OWLConcepts.toString());
        return abox;
    }

	public int Den(ATermAppl a){
		return ts.indexOf(a);
	}

	protected void applyARule(/*Individual ind*/){
		for(int i = 0; i < ts.getMax(); i++){
			if (ts.get(i) !=null) {

              Individual cur_node = abox.getIndividual((ATermAppl) ts.get(i));
              if (cur_node !=null) {
                if (ABox.DEBUG) System.out.println("Cur_node "+cur_node.debugString());
                if (i==ts.getMax()) {
                        return;
                } else {
                    if (ABox.DEBUG)
                        System.out.println("Next_node " + i + " " + abox.getIndividual((ATermAppl) ts.get(i+1)));
                        EdgeList ed = cur_node.getOutEdges();
//TODO check and if no temporal edges are found - create "relat" edges
                        EdgeList edcopy = new EdgeList();
                        int edgesize = ed.size();
                        if (edgesize > 0) {
                                if (!ed.hasEdgeTo(abox.getIndividual((ATermAppl)ts.get(i+1))) && (abox.getIndividual((ATermAppl)ts.get(i+1))!=null)) {
                                    cur_node.addEdge(abox.relat, abox.getIndividual((ATermAppl)ts.get(i+1)), DependencySet.INDEPENDENT);
//                                    System.out.println("ARule1: Addition of edge _relat_ between node " + cur_node.toString() + " and  node " + abox.getIndividual((ATermAppl)ts.get(i+1)).toString());
                                } else if (!ed.hasEdgeTo(abox.getIndividual((ATermAppl)ts.get(i+1))) && (abox.getIndividual((ATermAppl)ts.get(i+1))==null)) {
                                    abox.addIndividual((ATermAppl)ts.get(i+1));
                                    cur_node.addEdge(abox.relat, abox.getIndividual((ATermAppl)ts.get(i+1)), DependencySet.INDEPENDENT);
//                                    System.out.println("ARule2: Addition of edge _relat_ between node " + cur_node.toString() + " and  node " + abox.getIndividual((ATermAppl)ts.get(i+1)).toString());
                                }
                        } else {//if edgesize=0 - means no edges yet inducted on ts - induct them now
                            if (/*!ed.hasEdgeTo(abox.getIndividual((ATermAppl)ts.get(i+1))) && */(abox.getIndividual((ATermAppl)ts.get(i+1))!=null)) {
                                cur_node.addEdge(abox.relat, abox.getIndividual((ATermAppl)ts.get(i+1)), DependencySet.INDEPENDENT);
//                                System.out.println("ARule3: Addition of edge _relat_ between node " + cur_node.toString() + " and  node " + abox.getIndividual((ATermAppl)ts.get(i+1)).toString());

                            } else if (/*!ed.hasEdgeTo(abox.getIndividual((ATermAppl)ts.get(i+1))) && */(abox.getIndividual((ATermAppl)ts.get(i+1))==null)) {
                                abox.addIndividual((ATermAppl)ts.get(i+1));
                                cur_node.addEdge(abox.relat, abox.getIndividual((ATermAppl)ts.get(i+1)), DependencySet.INDEPENDENT);
//                                System.out.println("ARule4: Addition of edge _relat_ between node " + cur_node.toString() + " and  node " + abox.getIndividual((ATermAppl)ts.get(i+1)).toString());
                            }
                    }
                  }
	        }
            }
            }
        tsInducted = ts; //TODO to keep all timeInstants
    }

    protected void applyAtARule(Individual node) {
        if (!node.canApply(Node.AT)) return;
        List types = node.getTypes(Node.AT);
        int size = types.size();
        for (int j = node.applyNext[Node.AT]; j < size; j++) {
            //TODO create new nodes for "AT"
            ATermAppl c = (ATermAppl)types.get(j);
            try {
            int moment = Den((ATermAppl)c.getArgument(1));
            } catch (IntMapException e) {
                System.out.println("Application of AT operator over empty TimeStructure is not correct. Please check the TimeStructure and the Instant referenced in AT.");
            }
                if (abox.getIndividual((ATermAppl)c.getArgument(1)) != null) {
                    addType(abox.getIndividual((ATermAppl)c.getArgument(1)), (ATermAppl)c.getArgument(0), DependencySet.INDEPENDENT);
//TODO12/05/2009 add also an "relat" edge if there is no such edge from node to abox.getIndividual((ATermAppl)c.getArgument(1))
                    EdgeList ed = node.getOutEdges();
                    if (node.getOutEdges().size()==0) {
                        node.addEdge(abox.relat,abox.getIndividual((ATermAppl)c.getArgument(1)),DependencySet.INDEPENDENT);
                        if (ABox.DEBUG)
                            System.out.println("AtARule: Addition of edge _relat_ between node " + node.toString() + " and  node " + abox.getIndividual((ATermAppl)c.getArgument(1)).toString());

                    } else {
                        EdgeList edcopy = new EdgeList();
                        edcopy.addEdgeList(ed);
                        Iterator it = edcopy.iterator();
                        while (it.hasNext()) {
                            Edge edge = (Edge)it.next();
                            if ((edge.getRole()==abox.relat) && (edge.getTo()!= abox.getIndividual((ATermAppl)c.getArgument(1)))) {
                                node.addEdge(abox.relat,abox.getIndividual((ATermAppl)c.getArgument(1)),DependencySet.INDEPENDENT);
                                System.out.println("AtARule: Addition of edge _relat_ between node " + node.toString() + " and  node " + abox.getIndividual((ATermAppl)c.getArgument(1)).toString());
                            } else continue;
                        }
                    }
                } else {
                    Individual new_time_moment = abox.addIndividual((ATermAppl)c.getArgument(1));
                    new_time_moment.addType((ATermAppl)c.getArgument(0),DependencySet.INDEPENDENT);
//TODO12/05/2009 add edge "relat",as new_time_moment node is just created
                    node.addEdge(abox.relat, new_time_moment, DependencySet.INDEPENDENT);
                        if (ABox.DEBUG)
                            System.out.println("AtARule: Addition of edge _relat_ between node " + node.toString() + " and  node " + abox.getIndividual((ATermAppl)c.getArgument(1)).toString());
                }
        }
        node.applyNext[Node.AT]=size;
        return;
    }

    protected void applyFnRule(Individual node){
        if (!node.canApply(Node.FUT)) return;
        List types = node.getTypes(Node.FUT);
        int size = types.size();
        for(int j = node.applyNext[Node.FUT]; j < size; j++) {
            ATermAppl c = (ATermAppl) types.get(j);
            c = ATermUtils.nnf(c);
            int n = ((ATermInt) c.getArgument(1)).getInt();
            EdgeList ed = node.getOutEdges();
            EdgeList edcopy = new EdgeList();
            int edgesize = ed.size();
            if (edgesize > 0) {
                edcopy.addEdgeList(ed);
                Iterator it = edcopy.iterator();
                while(it.hasNext()){
                    Edge edge = (Edge)it.next();
                    if (ABox.DEBUG) System.out.println("Edge "+ edge);
                    if ((edge.getRole()==abox.relat) || (edge.getRole()==abox.relf)) {
                        Individual to = (Individual) edge.getTo();
                        if(n>1){
                            ATermAppl new_c = ATermUtils.makeFuture(c.getArgument(0),n-1);
                            to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                            if (edge.getRole() !=abox.relf){
                            node.addEdge(abox.relf, to, DependencySet.DUMMY);}
                            abox.changed = true;
                        }
                        else if(n==1){
                        ATermAppl new_c = (ATermAppl) c.getArgument(0);
                            to.addType(new_c,DependencySet.DUMMY);//TODO  look at ds
                            if (edge.getRole() !=abox.relf){
                            node.addEdge(abox.relf, to, DependencySet.DUMMY);}
                            abox.changed = true;
                        }
                    }
                }
            } else { //TODO it means no at_ti was in formula and no other edges<=> Time moment "now", only Fn/Pn and no other timeInstant-forming constructs
                Individual to = abox.addFreshTemporalIndividual();
                if(n>1){
                    ATermAppl new_c = ATermUtils.makeFuture(c.getArgument(0),n-1);
                    to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                    node.addEdge(abox.relf, to, DependencySet.DUMMY);
                    abox.changed = true;
                }
                else if(n==1){
                ATermAppl new_c = (ATermAppl) c.getArgument(0);
                    to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                    node.addEdge(abox.relf, to, DependencySet.DUMMY);
                    abox.changed = true;
                }
            }
        }
        node.applyNext[Node.FUT]=size;

        return;
    }

    protected void applyPnRule(Individual node){
        if (!node.canApply(Node.PAST)) return;
        List types = node.getTypes(Node.PAST);
        int size = types.size();
        for(int j = node.applyNext[Node.PAST]; j < size; j++) {
            ATermAppl c = (ATermAppl) types.get(j);
            c = ATermUtils.nnf(c);
            int n = ((ATermInt) c.getArgument(1)).getInt();
            EdgeList ed = node.getOutEdges();
            EdgeList edcopy = new EdgeList();
            int edgesize = ed.size();
            if (edgesize > 0) {
                edcopy.addEdgeList(ed);
                Iterator it = edcopy.iterator();
                while(it.hasNext()){
                    Edge edge = (Edge)it.next();
                    if (ABox.DEBUG) System.out.println("Edge "+ edge);
                    if ((edge.getRole()==abox.relat)|| (edge.getRole()==abox.relp)) {
                        Individual to = (Individual) edge.getTo();
                        if(n>1){
                            ATermAppl new_c = ATermUtils.makePast(c.getArgument(0),n-1);
                            to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                            if (edge.getRole()!=abox.relp) {
                            node.addEdge(abox.relp, to, DependencySet.DUMMY);}
                            abox.changed = true;
                        }
                        else if(n==1){
                            ATermAppl new_c = (ATermAppl) c.getArgument(0);
                            to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                            if (edge.getRole()!=abox.relp) {
                            node.addEdge(abox.relp, to, DependencySet.DUMMY);}
                            abox.changed = true;
                        }
                    }
                }
            } else {
                   Individual to = abox.addFreshTemporalIndividual();
                if(n>1){
                    ATermAppl new_c = ATermUtils.makePast(c.getArgument(0),n-1);
                    to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                    node.addEdge(abox.relp, to, DependencySet.DUMMY);
                    abox.changed = true;
                }
                else if(n==1){
                    ATermAppl new_c = (ATermAppl) c.getArgument(0);
                    to.addType(new_c,DependencySet.DUMMY);//TODO look at ds
                    node.addEdge(abox.relp, to, DependencySet.DUMMY);
                    abox.changed = true;
                }
            }
        }
        node.applyNext[Node.PAST]=size;
        return;
    }

	protected void applySomefutureRule(Individual node){
        if (!node.canApply(Node.SOMEFUT) || blocking.isBlocked(node)) return;
        List types = node.getTypes(Node.SOMEFUT);
        int size = types.size();
        for(int j = node.applyNext[Node.SOMEFUT]; j < size; j++) {
             ATermAppl c = (ATermAppl) types.get(j);
             c = ATermUtils.nnf(c);
             if (ABox.DEBUG) System.out.println("ApplySomefutureRule. Total types SOMEFUT: "+ size + ". Current element number is " + j + ". Current concept is " + c+ ".");
//TODO  is the node is not marked - create two branches
//TODO  branch 1: x:C add to types of node
//TODO  branch 2: x:future1 somefuture C add to types of node
			if(!node.isMarked()) {
                ATermAppl[] possibilities = new ATermAppl[2];
                possibilities[0] = (ATermAppl)c.getArgument(0);
                possibilities[1] = ATermUtils.makeFuture(c,1);

                    for(int k = 0, n = possibilities.length; k < n; k++) {
//TODO  Absolutely stupid way of reuse of code. Sorry. Shame on me.
                        ATermAppl a = (ATermAppl) possibilities[k];
                        ATermAppl[] poss = new ATermAppl[1];
                        poss[0] = a;
                        DisjunctionBranch newBranch = new DisjunctionBranch(abox, this, node,
                        c, DependencySet.INDEPENDENT, poss);

                        addBranch(newBranch);
                if(newBranch.tryNext() == false) return;
                }
            }
            else {//Should not work, if the domain is expanding!
//TODO  is the node is marked: create new individual new_node, copy all from node to new_node (copy (S,x,v))
//TODO  add type future1 somefuture C to new_node
//TODO  add type x:C to node
//TODO  make both nodes marked
                Individual new_node = abox.addFreshTemporalIndividual();
                new_node = (Individual)node.copyTo(abox);
                if (ABox.DEBUG)
                    System.out.println("New marked node according to somefut rule: " + new_node.debugString());
                ATermAppl new_c = ATermUtils.makeFuture(c,1);
                new_node.addType(new_c, DependencySet.DUMMY);
                node.addType(c, DependencySet.DUMMY);
                node.isMarked = true;
                new_node.isMarked = true;
			}
		}
        node.applyNext[Node.SOMEFUT]=size;
	 }

	protected void applySomepastRule(Individual node){
        if (!node.canApply(Node.SOMEPAST) || blocking.isBlocked(node)) return;
        List types = node.getTypes(Node.SOMEPAST);
        int size = types.size();
        for(int j = node.applyNext[Node.SOMEPAST]; j < size; j++) {
             ATermAppl c = (ATermAppl) types.get(j);
             c = ATermUtils.nnf(c);
             if (ABox.DEBUG) System.out.println("ApplySomepastRule. Total types SOMEPAsT: "+ size + ". Current element number is " + j + ". Current concept is " + c+ ".");
//TODO  is the node is not marked - create two branches
//TODO  branch 1: x:C add to types of node
//TODO  branch 2: x:past1 somepast C add to types of node
			if(!node.isMarked()) {
                ATermAppl[] possibilities = new ATermAppl[2];
                possibilities[0] = (ATermAppl)c.getArgument(0);
                possibilities[1] = ATermUtils.makePast(c,1);

                    for(int k = 0, n = possibilities.length; k < n; k++) {
//TODO Absolutely stupid way of reuse of code. Sorry. Shame on me.
                        ATermAppl a = (ATermAppl) possibilities[k];
                        ATermAppl[] poss = new ATermAppl[1];
                        poss[0] = a;
                        DisjunctionBranch newBranch = new DisjunctionBranch(abox, this, node,
                        c, DependencySet.INDEPENDENT, poss);

                        addBranch(newBranch);

                if(newBranch.tryNext() == false) return;
                }
            }
            else {//Should not work, if the domain is expanding!
//TODO  is the node is marked: create new individual new_node, copy all from node to new_node (copy (S,x,v))
//TODO  add type past1 somepast C to new_node
//TODO  add type x:C to node
//TODO  make both nodes marked
                Individual new_node = abox.addFreshTemporalIndividual();
                new_node = (Individual)node.copyTo(abox);
                if (ABox.DEBUG)
                    System.out.println("New marked node according to somepast rule: " + new_node.debugString());
                ATermAppl new_c = ATermUtils.makePast(c,1);
                new_node.addType(new_c, DependencySet.DUMMY);
                node.addType(c, DependencySet.DUMMY);
                node.isMarked = true;
                new_node.isMarked = true;
			}
		}
        node.applyNext[Node.SOMEPAST]=size;

	}

    protected void applyAllfutureRule(Individual node){
        if( !node.canApply(Node.ALLFUT)  || blocking.isBlocked(node)) return;
        List types = node.getTypes(Node.ALLFUT);
        int size = types.size();
        for(int j = node.applyNext[Node.ALLFUT]; j < size; j++) {
            ATermAppl c = (ATermAppl) types.get(j);
            ATermAppl c2 = ATermUtils.makeFuture(c,1);
            ATermAppl c3 = (ATermAppl)(c.getArgument(0));
            node.addType(c2, DependencySet.INDEPENDENT);
            node.addType(c3, DependencySet.INDEPENDENT);
            }
            node.applyNext[Node.ALLFUT]=size;
    }

    protected void applyAllpastRule(Individual node){
        if( !node.canApply(Node.ALLPAST) || blocking.isBlocked(node)) return;
        List types = node.getTypes(Node.ALLPAST);
        int size = types.size();
        for(int j = node.applyNext[Node.ALLPAST]; j < size; j++) {
            ATermAppl c = (ATermAppl) types.get(j);
            ATermAppl c2 = ATermUtils.makePast(c,1);
            ATermAppl c3 = (ATermAppl)(c.getArgument(0));
            node.addType(c2, DependencySet.INDEPENDENT);
            node.addType(c3, DependencySet.INDEPENDENT);
//             expand(node);
            }
        node.applyNext[Node.ALLPAST]=size;

    }

    private void expand(Individual x) {
/* TODO
if( tblocking.isFutureBlocked(x) && tblocking.isPastBlocked(x)) {
            mayNeedExpanding.removeFirst();
            return;
        }
*/
        if( blocking.isBlocked(x) ) {
            mayNeedExpanding.removeFirst();
            return;
        }

       do {
            applyUnfoldingRule(x);
            if(abox.isClosed()) return;
            applyDisjunctionRule(x);
            if(abox.isClosed()) return;
            applyARule(/*x*/);
            applyAtARule(x);
            applyFnRule(x);
            applyPnRule(x);
            applySomefutureRule(x);
            applySomepastRule(x);
            applyAllfutureRule(x);
            applyAllpastRule(x);
            if(abox.isClosed()) return;
            if (ABox.DEBUG){
                System.out.println("The state of the individual  "+ x +" is "+ x.debugString());}
       }
        while (x.canApply(Node.ATOM) || x.canApply(Node.OR) || x.canApply(Node.FUT) || x.canApply(Node.ALLFUT) || x.canApply(Node.PAST) || x.canApply(Node.ALLPAST));

        mayNeedExpanding.removeFirst();

// TODO random behaviour while choose the next individual for expansion - PelletOptions.DEPTH_FIRST - removed and now works nicely
//        int insert = ( PelletOptions.SEARCH_TYPE == PelletOptions.DEPTH_FIRST )
//        	? 0
//        	: mayNeedExpanding.size();
        int insert = mayNeedExpanding.size();

        mayNeedExpanding.addAll( insert, x.getSortedSuccessors() );
        if (ABox.DEBUG) System.out.println("The state of mayNeedExpanding list is:   "+ mayNeedExpanding.toString());

    }

    boolean supportsPseudoModelCompletion() {
        return true;
    }

//TODO copied from EmptySHNStrategy
    protected void applyDisjunctionRule(Individual node) {
        if(!node.canApply(Node.OR)) return;

        List types = node.getTypes(Node.OR);
        int size = types.size();
        ATermAppl[] disjunctions = new ATermAppl[size - node.applyNext[Node.OR]];
        types.subList(node.applyNext[Node.OR], size).toArray( disjunctions );
	    if( PelletOptions.USE_DISJUNCTION_SORTING != PelletOptions.NO_SORTING )
	        DisjunctionSorting.sort( node, disjunctions );

        for(int j = 0, n = disjunctions.length; j < n; j++) {
            ATermAppl disjunction = disjunctions[j];

            // disjunction is now in the form not(and([not(d1), not(d2), ...]))
            ATermAppl a = (ATermAppl) disjunction.getArgument(0);
            ATermList disjuncts = (ATermList) a.getArgument(0);
            ATermAppl[] disj = new ATermAppl[disjuncts.getLength()];

            for(int index = 0; !disjuncts.isEmpty(); disjuncts = disjuncts.getNext(), index++) {
                disj[index] = ATermUtils.negate( (ATermAppl) disjuncts.getFirst() );
                if(node.hasType(disj[index])) break;
            }

            if(!disjuncts.isEmpty()) continue;

            DisjunctionBranch newBranch = new DisjunctionBranch(abox, this, node,
                disjunction, node.getDepends(disjunction), disj);

            addBranch(newBranch);

            if(newBranch.tryNext() == false) return;
        }
        node.applyNext[Node.OR] = size;

    }
//=======================getNextIndividual-taken from EmptySHNStrategy=========
    private Individual getNextIndividual() {
    Node next = null;
        while( !mayNeedExpanding.isEmpty() ) {
            next = (Node) mayNeedExpanding.getFirst();
            if( next instanceof Literal ) {
                next = null;
                mayNeedExpanding.removeFirst();
            }
            else
                break;
        }

        return (Individual) next;
    }

    public ATermAppl createOWLConcept(Individual node) {
        ATermAppl result = null;
//        if (!node.canApply(Node.AT)) return ATermUtils.TOP;
        //for (int i = 0;i<Node.NOM;i++) {
            Set types = new HashSet( node.getTypes() );
            types.remove( ATermUtils.TOP );
            types.remove( ATermUtils.makeValue( node.getName() ) );
            int count = 0;
            ATerm[] terms = new ATerm[ types.size() ];
            for(Iterator it = types.iterator(); it.hasNext(); ) {
                ATermAppl c = (ATermAppl) it.next();
                if( !ATermUtils.isAnd(c) && !ATermUtils.isAt(c) &&
                        !ATermUtils.isAllfuture(c) && !ATermUtils.isSomefuture(c) && !ATermUtils.isFuture(c) &&
                        !ATermUtils.isAllpast(c) && !ATermUtils.isSomepast(c) && !ATermUtils.isPast(c))
                    {
                            terms [count++] = replaceBnodes(c);
                    }
            }
            if (count>0)
                result = ATermUtils.makeAnd( ATermUtils.toSet( terms, count ) );
    return result;
    }

    public ATermAppl createMomentConcept(Individual node) {
        ATermAppl result = null;
        Set types = new HashSet( node.getTypes() );
        if (node.getTypes(Node.NOM).contains(ATermUtils.makeValue( node.getName() ) )) {
        result = node.getName();
        return result;}
        else  return ATermUtils.makeTermAppl("now");
    }

    public ATermAppl replaceBnodes (ATermAppl c) {
        ATermAppl new_c = c;
        if (ATermUtils.isBnode(c)) {
            new_c = (ATermAppl) unfoldingMap.get(c);
            AFun fun = new_c.getAFun();
        
            if (fun.equals(ATermUtils.NOTFUN)) {
                ATerm arg1 = new_c.getArgument(0);
                new_c = replaceBnodes ((ATermAppl)arg1);
                new_c = ATermUtils.makeNot(new_c);
            }
            if (fun.equals(ATermUtils.ANDFUN)) {
                ATerm arg1 = new_c.getArgument(0);
                ATerm arg2 = new_c.getArgument(1);

                ATerm new_arg1 = replaceBnodes((ATermAppl)arg1);
                ATerm new_arg2 = replaceBnodes((ATermAppl)arg2);
                new_c = (ATermUtils.makeAnd(new_arg1,new_arg2));
            }
        } else if (ATermUtils.isAnd(c)) {
            ATerm arg1 = c.getArgument(0);
            if (arg1.getType() == ATerm.LIST) {
                ATermList new_arg1 = replaceBnodes((ATermList)arg1);
                new_c = (ATermUtils.makeAnd(new_arg1));
            } else if (arg1.getType() == ATerm.APPL) {
                ATermAppl new_arg1 = replaceBnodes((ATermAppl) arg1);
                ATerm arg2 = c.getArgument(1);
                ATerm new_arg2 = replaceBnodes((ATermAppl)arg2);
                new_c = (ATermUtils.makeAnd(new_arg1,new_arg2));
            }
        } else if (ATermUtils.isNot(c)) {
            ATerm arg1 = c.getArgument(0);
            new_c = replaceBnodes ((ATermAppl)arg1);
            new_c = ATermUtils.makeNot(new_c);
        }
       return new_c;
    }

    public ATermList replaceBnodes (ATermList list) {
        ATermAppl new_c;
        ATermList new_list = ATermUtils.EMPTY_LIST;
        for(; !list.isEmpty(); list = list.getNext()) {
            ATermAppl conj = (ATermAppl) list.getFirst();
            new_c = replaceBnodes(conj);
            new_list = new_list.insert(ATermUtils.nnf(new_c));
        }
    return new_list;
    }

}
