//
// Copyright (c) 2003 Ron Alford, Mike Grove, Bijan Parsia, Evren Sirin
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

package org.mindswap.pellet.jena;

import com.hp.hpl.jena.graph.*;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.rdf.model.impl.*;
import com.hp.hpl.jena.reasoner.*;
import com.hp.hpl.jena.vocabulary.*;

/**
 * User: NataKeberle
 * Date: 16/5/2009
 * Time: 14:03:48
 */
public class PelletMetReasoner implements Reasoner {

    private PelletMetReasonerFactory factory;

    private Capabilities graphCapabilities;

    private Graph schema;

    public PelletMetReasoner( ) {
        this( null, PelletMetReasonerFactory.theInstance() );
    }

    public PelletMetReasoner( PelletMetReasonerFactory factory ) {
        this( null, factory );
    }

    public PelletMetReasoner( Graph schema, PelletMetReasonerFactory factory ) {
        this.schema = schema;
        this.factory = factory;

        graphCapabilities = new BaseInfGraph.InfCapabilities();
     }

    public Graph getSchema() {
        return schema;
    }

    public Reasoner bindSchema( Graph graph ) throws ReasonerException {
        return new org.mindswap.pellet.jena.PelletMetReasoner( graph, factory );
    }

    public Reasoner bindSchema( Model model ) throws ReasonerException {
        return bindSchema( model.getGraph() );
    }

    public InfGraph bind( Graph graph ) throws ReasonerException {
        return new PelletMetInfGraph( graph, this );
    }

    public InfModel bind( Model model ) throws ReasonerException {
    	return new InfModelImpl( bind( model.getGraph() ) );
    }

    public void setDerivationLogging( boolean enable ) {
    }

    public void setParameter(Property arg0, Object arg1) {
    }

    public Model getReasonerCapabilities() {
        return (factory == null) ? null : factory.getCapabilities();
    }

    public Capabilities getGraphCapabilities() {
        return graphCapabilities;
    }

    public void addDescription(Model arg0, Resource arg1) {
    }

    public boolean supportsProperty(Property property) {
        if( factory == null ) return false;
        Model caps = factory.getCapabilities();
        Resource root = caps.getResource(factory.getURI());
        return caps.contains(root, ReasonerVocabulary.supportsP, property);
    }

}
