package org.mindswap.pelletmet;

/**
 * User: NataKeberle
 * Date: 23/07/2009
 * Time: 20:39:31
 */

import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.rdf.model.*;
import org.mindswap.pellet.*;
import org.mindswap.pellet.jena.*;
import org.mindswap.pellet.utils.*;

import java.io.*;

public class OWLMeTTest {

    public static boolean DEBUG = false;
    public static boolean PRINT_TIME = false;
    public static boolean PRINT_TREE = false;

    // time limits for different kind of tests
    public static int SAT_LIMIT  = 10;
    public static int TBOX_LIMIT = 100;
    public static int ABOX_LIMIT = 200;

    public static boolean FAST = false;
    public static boolean FORCE_UPPERCASE = true;

    private KnowledgeBase kb;
    private ModelReader modelReader;
    private OntModel model;
	private OWLMETReasoner reasoner;

    public OWLMeTTest() {
	    modelReader = new ModelReader();
        reasoner = new OWLMETReasoner();
	}

	public KnowledgeBase initKB(long timeout) {
	    KnowledgeBase kb = new KnowledgeBase();
		kb.setTimeout(timeout * 1000);

		return kb;
	}

    public boolean doTBoxTest( String file ) throws Exception {
        String ext = ".txt";
        int index = file.lastIndexOf('.');
        if(index != -1) {
            ext = file.substring( index );
            file = file.substring( 0, index );
        }
        index = file.lastIndexOf(File.separator);
        String displayName = (index == -1) ? file : file.substring(index + 1);
        System.out.print(displayName + " ");

        kb = initKB(TBOX_LIMIT);
        Model plainModel = null;
        plainModel = modelReader.read(file+ext, "RDF/XML");
        if (plainModel != null) {
            model = ModelFactory.createOntologyModel( PelletMetReasonerFactory.THE_SPEC, plainModel );
            reasoner = ((PelletMetInfGraph) model.getGraph()).getOWLMETReasoner();
        }
        Timer t = kb.timers.startTimer( "test" );
        ((PelletMetInfGraph) model.getGraph()).setLazyConsistency( true );

        model.prepare();
        kb = reasoner.getKB();

        kb.classify();
        Boolean consistent = kb.isConsistent();

        t.stop();

        if(PRINT_TREE) kb.printClassTree();

        System.out.print( " Consistent? " +  consistent + "\n");
        System.out.print( " Prepare " + kb.timers.getTimer("preprocessing").getTotal() );
        System.out.print( " Classify " + kb.timers.getTimer("classify").getTotal() );
        System.out.print( "\n TOTAL TIME " + t.getTotal() + "\n\n");
        if(PRINT_TIME) kb.timers.print();

        return true;
    }
}
